
signature LOGIC =
   sig
      
      datatype prop =
         Atomic of string
       | And of prop * prop
       | Implies of prop * prop
       | Or of prop * prop
       | True
       | False

      val propToString : prop -> string

      (* The parser requires atomic propositions to have single-character, alphanumeric names *)
      exception SyntaxError
      val parseProp : string -> prop

   end


structure Logic :> LOGIC =
   struct

      datatype prop =
         Atomic of string
       | And of prop * prop
       | Implies of prop * prop
       | Or of prop * prop
       | True
       | False

      val precAnd = 3
      val precOr = 2
      val precImplies = 1
      val precMin = 0

      fun left b =
         if b then
            ["("]
         else
            []

      fun right b =
         if b then
            [")"]
         else
            []

      fun propToStringMain prec p suffix =
         (case p of
             Atomic str =>
                str :: suffix

           | And (p1, p2) =>
                left (prec > precAnd)
                @
                propToStringMain (precAnd+1) p1
                   (" /\\ "
                    ::
                    propToStringMain precAnd p2
                       (right (prec > precAnd) @ suffix))

           | Or (p1, p2) =>
                left (prec > precOr)
                @
                propToStringMain (precOr+1) p1
                   (" \\/ "
                    ::
                    propToStringMain precOr p2
                       (right (prec > precOr) @ suffix))

           | Implies (p1, p2) =>
                left (prec > precImplies)
                @
                propToStringMain (precImplies+1) p1
                   (" => "
                    ::
                    propToStringMain precImplies p2
                       (right (prec > precImplies) @ suffix))

           | True =>
                "T" :: suffix

           | False =>
                "F" :: suffix)

      fun propToString p = String.concat (propToStringMain precMin p [])


      datatype element = Oper of (prop * prop -> prop) * int | Term of prop | Paren

      exception SyntaxError

      fun flush prec stack =
         (case stack of
             [Term _] => stack

           | Term _ :: Paren :: _ => stack

           | Term q :: Oper (f, prec') :: Term p :: stack' =>
                if prec >= prec' then
                   stack
                else
                   flush prec (Term (f (p, q)) :: stack')

           | _ => raise SyntaxError)

      fun ready stack =
         (case stack of
             [] => stack
           | Paren :: _ => stack
           | Oper _ :: _ => stack
           | _ => raise SyntaxError)

      fun parsePropMain stack l =
         (case l of
             [] =>
                (case flush precMin stack of
                    [Term p] => (p, [])

                  | _ => raise SyntaxError)

           | #"/" :: #"\\" :: rest =>
                parsePropMain (Oper (And, precAnd) :: flush precAnd stack) rest

           | #"\\" :: #"/" :: rest =>
                parsePropMain (Oper (Or, precOr) :: flush precOr stack) rest

           | #"=" :: #">" :: rest =>
                parsePropMain (Oper (Implies, precImplies) :: flush precImplies stack) rest

           | #"T" :: rest =>
                parsePropMain (Term True :: ready stack) rest

           | #"F" :: rest =>
                parsePropMain (Term False :: ready stack) rest

           | #"(" :: rest =>
                parsePropMain (Paren :: ready stack) rest

           | #")" :: rest =>
                (case flush precMin stack of
                    Term p :: Paren :: stack' =>
                       parsePropMain (Term p :: stack') rest

                  | _ => raise SyntaxError)

           | ch :: rest =>
                if Char.isSpace ch then
                   parsePropMain stack rest
                else if Char.isAlphaNum ch then
                   parsePropMain (Term (Atomic (String.str ch)) :: ready stack) rest
                else
                   (case flush precMin stack of
                       [Term p] => (p, l)

                     | _ => raise SyntaxError))

      fun parseProp str =
         (case parsePropMain [] (String.explode str) of
             (p, []) => p

           | _ => raise SyntaxError)

   end
